// Copyright (c) 1994,1995 James Clark
// See the file COPYING for copying permission.

#include "config.h"
#include "SgmlParser.H"
#include "Event.H"
#include "MessageEventHandler.H"
#include "SgmlsEventHandler.H"
#include "RastEventHandler.H"
#include "OutputCharStream.H"
#include "EasyParser.H"
#include "Boolean.H"
#include "programName.H"
#include "getopt.h"
#include "macros.H"
#include "NsgmlsMessages.H"
#include "MessageArg.H"
#include "ErrnoMessageArg.H"

#include "outOfMemory.H"
#include "xnew.h"

#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#ifdef HAVE_LOCALE
#include <locale.h>
#endif

inline
void usage(Messenger &mgr)
{
  mgr.message(NsgmlsMessages::usageInfo);
}

static int parseProlog(EasyParser &easy, unsigned flags);
static int validate(EasyParser &easy, unsigned flags);
static int outputSgmls(EasyParser &easy, streambuf *, Boolean outputLocations);
static int outputRast(EasyParser &easy, const char *filename);

static Boolean exotericaMode = 0; // hack, hack

int main(int argc, char **argv)
{
  set_new_handler(outOfMemory);
#ifdef HAVE_LOCALE
  setlocale(LC_ALL, "");
#endif
  programName = argv[0];
  int opt;
  Boolean suppressOutput = 0;
  EasyParser easy(getenv("NSGMLS_CODE"));
  Boolean prologOnly = 0;
  Boolean lFlag = 0;
  const char *errorFile = 0;
  const char *rastFile = 0;
  char arg[2];

  opterr = 0;			// no messages from getopt
  // Colon at beginning is Posix.2ism that says to return : rather than ? for
  // missing option argument.
  while ((opt = getopt(argc, argv, ":a:c:def:gi:lm:prst:uvw:xX")) >= 0) {
    switch (opt) {
    case 'a':
      // activate link
      easy.activateLinkType(optarg);
      break;
    case 'c':
      // output RACT file
      easy.message(NsgmlsMessages::noRactOption);
      break;
    case 'd':
      // warn about duplicate entity declarations
      easy.options.warnDuplicateEntity = 1;
      break;
    case 'e':
      // describe open entities in error messages
      easy.addOption(MessageReporter::openEntities);
      break;
    case 'f':
      errorFile = optarg;
      break;
    case 'g':
      // show gis of open elements in error messages
      easy.addOption(MessageReporter::openElements);
      break;
    case 'i':
      // pretend that optarg is defined as INCLUDE
      easy.addInclude(optarg);
      break;
    case 'l':
      // output L commands
      lFlag = 1;
      break;
    case 'm':
      easy.addCatalogSysid(optarg);
      break;
    case 'p':
      prologOnly = 1;
      break;
    case 'r':
      // warn about defaulted entity reference
      easy.options.warnDefaultEntityReference = 1;
      break;
    case 's':
      suppressOutput = 1;
      break;
    case 't':
      rastFile = optarg;
      break;
    case 'u':
      // warn about undefined elements
      easy.options.warnUndefinedElement = 1;
      break;
    case 'v':
      // print the version number
      {
	extern const char *versionString;
	// FIXME perhaps we should have a constant argument instead
	easy.message(NsgmlsMessages::versionInfo,
		     StringMessageArg(easy.codingSystem()
				      ->convertIn(versionString)));
      }
      break;
    case 'w':
      if (!easy.enableWarning(optarg))
	easy.message(NsgmlsMessages::unknownWarning,
		     StringMessageArg(easy.codingSystem()
				      ->convertIn(optarg)));
      break;
    case 'x':
      easy.options.inhibitIdrefCheck = 1;
      break;
    case 'X':
      exotericaMode = 1;
      break;
    case ':':
      arg[0] = optopt;
      arg[1] = '\0';
      easy.message(NsgmlsMessages::missingOptionArgError,
		   StringMessageArg(easy.codingSystem()->convertIn(arg)));
      usage(easy);
      return 1;
    case '?':
      arg[0] = optopt;
      arg[1] = '\0';
      easy.message(NsgmlsMessages::invalidOptionError,
		   StringMessageArg(easy.codingSystem()->convertIn(arg)));
      usage(easy);
      return 1;
    default:
      CANNOT_HAPPEN();
    }
  }
#ifdef __GNUG__
  // cout is a performance disaster in libg++ unless we do this.
  ios::sync_with_stdio(0);
#endif
  if (errorFile) {
    static filebuf file;
    if (!file.open(errorFile, ios::out|ios::trunc)) {
      easy.message(NsgmlsMessages::cannotOpenOutputError,
		   StringMessageArg(easy.codingSystem()
				    ->convertIn(errorFile)),
		   ErrnoMessageArg(errno));
      exit(1);
    }
    easy.setErrorStream(&file);
  }
  // could have option to parse all declared subdocuments
  easy.init(argc - optind, argv + optind, EasyParser::allLinkTypesActivated);
  if (prologOnly)
    return parseProlog(easy, 0);
  else if (rastFile)
    return outputRast(easy, rastFile);
  else if (suppressOutput)
    return validate(easy, MessageEventHandler::parseSubdocs);
  else 
    return outputSgmls(easy, cout.rdbuf(), lFlag);
}

static
int parseProlog(EasyParser &easy, unsigned flags)
{
  SgmlParser &parser = easy.parser();
  MessageEventHandler handler(&parser, &easy, flags);
  Event *event;
  while ((event = parser.nextEvent()) != 0
	 && event->type() != Event::endProlog)
    event->handle(handler);
  return handler.errorCount() > 0;
}

static
int validate(EasyParser &easy, unsigned flags)
{
  SgmlParser &parser = easy.parser();
  MessageEventHandler handler(&parser, &easy, flags);
  parser.parseAll(handler);
  return handler.errorCount() > 0;
}

static
int outputSgmls(EasyParser &easy, streambuf *stream, Boolean lFlag)
{
  IosOutputCharStream os(stream, easy.codingSystem());
  SgmlParser &parser = easy.parser();
  SgmlsEventHandler handler(&parser, &easy.entityManager(), &os, &easy, lFlag);
  parser.parseAll(handler);
  handler.endDocument();
  return handler.errorCount() > 0;
}

class XRastEventHandler : public RastEventHandler {
public:
  XRastEventHandler(SgmlParser *,
		    const char *filename,
		    const OutputCodingSystem *,
		    const InputCodingSystem *,
		    MessageReporter *reporter);
  void message(MessageEvent *);
  void truncateOutput();
private:
  MessageReporter *reporter_;
  // file_ must come before os_ so it gets inited first
  filebuf file_;
  IosOutputCharStream os_;
  const char *filenameBytes_;
  const InputCodingSystem *filenameCodingSystem_;
};

XRastEventHandler::XRastEventHandler(SgmlParser *parser,
				     const char *filenameBytes,
				     const OutputCodingSystem *codingSystem,
				     const InputCodingSystem *filenameCodingSystem,
				     MessageReporter *reporter)
: RastEventHandler(parser, reporter),
  reporter_(reporter),
  filenameBytes_(filenameBytes),
  filenameCodingSystem_(filenameCodingSystem)
{
  errno = 0;
  if (!file_.open(filenameBytes, ios::out|ios::trunc)) {
    reporter->message(NsgmlsMessages::cannotOpenOutputError,
		      StringMessageArg(filenameCodingSystem_
				       ->convertIn(filenameBytes_)),
		      ErrnoMessageArg(errno));
    exit(1);
  }
  os_.open(&file_, codingSystem);
  setOutputStream(&os_);
}

void XRastEventHandler::truncateOutput()
{
  os_.flush();
  errno = 0;
  if (!file_.close())
    reporter_->message(NsgmlsMessages::closeOutputError,
		       StringMessageArg(filenameCodingSystem_
					->convertIn(filenameBytes_)),
		       ErrnoMessageArg(errno));
  errno = 0;
  if (!file_.open(filenameBytes_, ios::out|ios::trunc)) {
    reporter_->message(NsgmlsMessages::cannotOpenOutputError,
		       StringMessageArg(filenameCodingSystem_
					->convertIn(filenameBytes_)),
		       ErrnoMessageArg(errno));
    exit(1);
  }
}

void XRastEventHandler::message(MessageEvent *event)
{
  static const char msg[] =
    "character %1 is not significant in the reference concrete syntax"
    " and so cannot occur in a literal in the SGML declaration"
    " except as the replacement of a character reference";
  // FIXME
  if (exotericaMode
      && strcmp(event->message().type->id, msg) == 0) {
    delete event;
    return;
  }
  reporter_->reportMessage(event->message());
  RastEventHandler::message(event);
}

static
int outputRast(EasyParser &easy, const char *filename)
{
  SgmlParser &parser = easy.parser();
  XRastEventHandler handler(&parser, filename, easy.codingSystem(),
			    easy.codingSystem(), &easy);
  parser.parseAll(handler);
  handler.end();
  return handler.hadError();
}
